import tensorflow.keras.losses as losses
from .weighted import weighted_categorical_crossentropy
#
# CONSTANTS
#
DEFAULT_LOSS='categorical_crossentropy'
DEFAULT_WEIGHTED_LOSS='weighted_categorical_crossentropy'


#
# LOSS FUNCTION DICT
#
LOSS_FUNCTIONS={
    'weighted_categorical_crossentropy': weighted_categorical_crossentropy,
    'categorical_crossentropy': losses.CategoricalCrossentropy
}


#
# MAIN
#
def get(loss_func=None,weights=None,**kwargs):
    if not loss_func:
        if weights:
            loss_func=DEFAULT_WEIGHTED_LOSS
            kwargs['weights']=weights
        else:
            loss_func=DEFAULT_LOSS
    if isinstance(loss_func,str):
        loss_func=LOSS_FUNCTIONS.get(loss_func,loss_func)
    if not isinstance(loss_func,str):
        loss_func=loss_func(**kwargs)
    return loss_func





import tensorflow as tf
import tfbox.utils.helpers as h


def get(metric):
    """ get keras metric by name
    Args:

        - metric<str|metric>:
            * snake-case strings will be turned to camel-case
            * if metric is not a string the passed metric will be returned
              with the assumption that it is already a keras metric class

    Return: metric class 
    """
    if isinstance(metric,str):
        metric=h.camel(metric)
        metric=getattr(tf.keras.metrics,metric)
    return metric


def weighted(weights,metric='categorical_accuracy'):
    """ weighted metric
    Args:

        - weights<list>:
            * a weight for each value in y_true
        - metric<str|metric>:
            * snake-case strings will be turned to camel-case
            * if metric is not a string the passed metric will be returned
              with the assumption that it is already a keras metric class

    Return: (weighted) metric instance 
    """
    metric=get(metric)()
    def _weighted_metric(y_true,y_pred):
        sample_weight=tf.reduce_sum(weights*y_true, axis=-1)
        metric.reset_states()
        metric.update_state(y_true,y_pred,sample_weight=sample_weight)
        return metric.result()
    _weighted_metric.name=f'{metric.name}-w'
    return _weighted_metric
 

def subset(ignore_labels,labels,metric='categorical_accuracy'):
    """ metric which ignores labels (wrapper for `weighted`)
    Args:

        - ignore_labels<list|int|float>:
            * label(s) to not include in metric calculation
        - labels<list|int|float>:
            * if (int|float) assumes labels are range(labels)
        - metric<str|metric>:
            * snake-case strings will be turned to camel-case
            * if metric is not a string the passed metric will be returned
              with the assumption that it is already a keras metric class

    Return: (weighted) metric instance 
    """
    if isinstance(ignore_labels,(int,float)):
        ignore_labels=[int(ignore_labels)]
    if isinstance(labels,(int,float)):
        labels=list(range(labels))
    weights=[int(i not in ignore_labels) for i in labels]
    return weighted(weights,metric=metric)
 





from pprint import pprint
import tensorflow.keras as keras
from . import load
from . import blocks
import tfbox.utils.helpers as h


# 
# Model: Parent class for TBox models/blocks
# 
# a simple wrapper of keras.Model with the following additions:
#
#   - an optional classifier
#   - is_skip property 
#   - standardized naming for tfbox models/blocks  
# 
class Model(keras.Model):
    #
    # CONSTANTS
    #
    NAME='TFBoxModel'
    DEFAULT_KEY=NAME
    SEGMENT='segment'
    GLOBAL_POOLING='global_pooling'


    def __init__(self,
            nb_classes=None,
            classifier_config={},
            classifier_key_path=None,
            classifier_is_file_path=False,
            classifier_cfig_dir=load.TFBOX,
            is_skip=False,  
            name=NAME,
            named_layers=True,
            noisy=True):
        super(Model, self).__init__()
        self.is_skip=is_skip
        self.model_name=name
        self.named_layers=named_layers
        # classifier
        if nb_classes:
            if isinstance(classifier_config,str):
                classifier_config=load.config(
                        cfig=classifier_config,
                        key_path=classifier_key_path,
                        is_file_path=classifier_is_file_path,
                        cfig_dir=classifier_cfig_dir,
                        noisy=noisy )
            classifier_type=classifier_config.pop('classifier_type')
            if classifier_type==Model.SEGMENT:
                self.classifier=blocks.SegmentClassifier(
                    nb_classes=nb_classes,
                    **classifier_config)
            elif classifier_type==Model.GLOBAL_POOLING:
                raise NotImplementedError('TODO: GAPClassifier')
            else:
                raise NotImplementedError(f'{classifier_type} is not a valid classifier')
        else:
            self.classifier=None


    def layer_name(self,group=None,index=None):
        return blocks.layer_name(self.model_name,group,index=index,named=self.named_layers)







from pprint import pprint
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from tfbox.utils.helpers import StrideManager
from . import base
from . import blocks
from . import load
#
#
# CONSTANTS
#




#
# Xception Network:
#
class Xception(tf.keras.Model):
    #
    # CONSTANTS
    #
    AUTO='auto'
    DEFAULT_KEY='xlight_os8_mf16'
    DEFAULT_SEGEMENT_KEY='xlight_os8_mf16_seg'
    DEFAULTS=load.config(cfig='xception',key_path=DEFAULT_KEY)
    GAP='gap'
    SEGMENT='segment'


    #
    # STATIC
    #
    @staticmethod
    def from_config(
            cfig='xception',
            key_path=DEFAULT_KEY,
            cfig_dir=load.TFBOX,
            is_file_path=False,
            **kwargs):
        config=load.config(
            cfig=cfig,
            key_path=key_path,
            cfig_dir=cfig_dir,
            is_file_path=is_file_path,
            **kwargs)
        print('XCEPTION:')
        pprint(config)
        return Xception(**config)


    #
    # PUBLIC
    #
    def __init__(self,
            output_stride=DEFAULTS['output_stride'],
            dropout=DEFAULTS.get('dropout',False),
            entry_flow_prestrides=DEFAULTS['entry_flow_prestrides'],
            entry_flow_prefilters_stack=DEFAULTS['entry_flow_prefilters_stack'],
            entry_flow_filters_stack=DEFAULTS['entry_flow_filters_stack'],
            entry_flow_strides_stack=DEFAULTS.get('entry_flow_strides_stack'),
            entry_flow_seperable=DEFAULTS.get('entry_flow_seperable',True),
            entry_flow_dropout=DEFAULTS.get('entry_flow_dropout',False),
            middle_flow_filters=DEFAULTS['middle_flow_filters'],
            middle_flow_depth=DEFAULTS['middle_flow_depth'],
            middle_flow_seperable=DEFAULTS.get('middle_flow_seperable',True),
            middle_flow_dropout=DEFAULTS.get('middle_flow_dropout',False),
            exit_flow_filters_in=DEFAULTS['exit_flow_filters_in'],
            exit_flow_filters=DEFAULTS['exit_flow_filters'],
            exit_flow_seperable=DEFAULTS.get('exit_flow_seperable',True),
            exit_flow_postfilters_stack=DEFAULTS['exit_flow_postfilters_stack'],
            exit_flow_dropout=DEFAULTS.get('exit_flow_dropout',False),
            nb_classes=DEFAULTS.get('nb_classes',None),
            classifier_type=DEFAULTS.get('classifier_type',False),
            classifier_act=DEFAULTS.get('classifier_act',True),
            classifier_act_config=DEFAULTS.get('classifier_act_config',{}),
            classifier_kernel_size_list=DEFAULTS.get('classifier_kernel_size_list'),
            classifier_filters_list=DEFAULTS.get('classifier_filters_list'),
            keep_mid_step=DEFAULTS.get('keep_mid_step',True),
            skip_indices=DEFAULTS.get('skip_indices',True),
            name=DEFAULTS.get('name',None),
            named_layers=DEFAULTS.get('named_layers',True)):
        super(Xception, self).__init__()
        self.model_name=name
        self.named_layers=named_layers
        if dropout:
           entry_flow_dropout=dropout
           middle_flow_dropout=dropout
           exit_flow_dropout=dropout


        self.stride_manager=StrideManager(
            output_stride,
            keep_mid_step=keep_mid_step,
            keep_indices=skip_indices)

        self.entry_stack, filters_out=self._entry_flow(
            entry_flow_prestrides,
            entry_flow_prefilters_stack,
            entry_flow_filters_stack,
            entry_flow_strides_stack,
            entry_flow_seperable,
            entry_flow_dropout)

        if middle_flow_filters:
            self.middle_stack, filters_out=self._middle_flow(
                middle_flow_filters,
                middle_flow_depth,
                middle_flow_seperable,
                middle_flow_dropout,
                filters_out)
        else:
            self.middle_stack=False
        if exit_flow_filters:
            self.exit_stack, filters_out=self._exit_flow(
                exit_flow_filters_in,
                exit_flow_filters,
                exit_flow_seperable,
                exit_flow_dropout,
                exit_flow_postfilters_stack,
                filters_out)
        else:
            self.exit_stack=False
        if classifier_type==Xception.GAP:
            # todo: global-avg-pooling+dense+classifier
            pass
        elif classifier_type==Xception.SEGMENT:
            self.classifier=blocks.SegmentClassifier(
                nb_classes=nb_classes,
                filters_list=classifier_filters_list,
                kernel_size_list=classifier_kernel_size_list,
                output_act=classifier_act,
                output_act_config=classifier_act_config,
                name=self._layer_name('classifier'),
                named_layers=self.named_layers)
        else:
            self.classifier=False



    def __call__(self,x,training=False,**kwargs):
        x,entry_skips=self._process_stack(self.entry_stack,x)
        if self.middle_stack:
            x=self._process_stack(self.middle_stack,x,False)
        if self.exit_stack:            
            x,exit_skips=self._process_stack(self.exit_stack,x)
        if self.classifier:
            return self.classifier(x)
        else:
            return x, entry_skips+exit_skips



    #
    # INTERNAL
    #
    def _layer_name(self,group=None,index=None):
        return blocks.layer_name(
            *[self.model_name,group],
            index=index,
            named=self.named_layers)


    def _entry_flow(self,prestrides,prefilters,filters,strides,seperable,dropout):
        _layers=[]
        for i,(s,f) in enumerate(zip(prestrides,prefilters)):
            _layers.append(blocks.Conv(
                filters=f,
                strides=self.stride_manager.strides(s),
                dilation_rate=self.stride_manager.dilation_rate,
                dropout=dropout,
                keep_output=self.stride_manager.keep_index,
                name=self._layer_name('entry',i),
                named_layers=self.named_layers))
            self.stride_manager.step(s)
        if not strides:
            strides=[2]*len(filters)
        for j,(f,s) in enumerate(zip(filters,strides)):
            _layers.append(blocks.Stack(
                    seperable=seperable,
                    dropout=dropout,
                    depth=3,
                    filters=f,
                    output_stride=self.stride_manager.strides(s),
                    dilation_rate=self.stride_manager.dilation_rate,
                    keep_output=self.stride_manager.keep_index,
                    name=self._layer_name('entry',i+j),
                    named_layers=self.named_layers))
            self.stride_manager.step(s)
        return _layers, filters[-1]


    def _middle_flow(self,filters,flow_depth,seperable,dropout,prev_filters):
        _layers=[]
        if filters==Xception.AUTO:
            filters=prev_filters
        for i in range(flow_depth):
            _layers.append(blocks.Stack(
                seperable=seperable,
                dropout=dropout,
                depth=3,
                filters=filters,
                dilation_rate=self.stride_manager.dilation_rate,
                residual=blocks.Stack.IDENTITY,
                name=self._layer_name('middle',i),
                named_layers=self.named_layers))
        return _layers, filters


    def _exit_flow(self,filters_in,filters,seperable,dropout,postfilters,prev_filters):
        if filters_in==Xception.AUTO:
            filters_in=prev_filters
        _layers=[]
        _layers.append(blocks.Stack(
                seperable=seperable,
                dropout=dropout,
                depth=3,
                filters=filters,
                filters_in=filters_in,
                output_stride=self.stride_manager.strides(),
                dilation_rate=self.stride_manager.dilation_rate,
                keep_output=False,
                name=self._layer_name('exit',0),
                named_layers=self.named_layers))
        self.stride_manager.step()
        for i,f in enumerate(postfilters):
            _layers.append(blocks.Conv(
                filters=f,
                seperable=seperable,
                dropout=dropout,
                dilation_rate=self.stride_manager.dilation_rate,
                name=self._layer_name('exit',i+1),
                named_layers=self.named_layers))
        if postfilters:
            filters=postfilters[-1]
        else:
            filters=filters[-1]
        return _layers, filters


    def _process_stack(self,stack,x,return_skips=True):
        if return_skips:
            skips=[]
        for layer in stack:
            x=layer(x)
            if return_skips:
                skips=self._update_skips(layer,skips,x)
        if return_skips:
            return x, skips
        else:
            return x


    def _update_skips(self,layer,skips,x,force_update=False):
        try:
            if force_update or (layer.keep_output): 
                skips.append(x)
        except:
            pass
        return skips



from pprint import pprint
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from tfbox.utils.helpers import StrideManager
from . import base
from . import blocks
from . import load
#
# Steps Network:
#
class Steps(tf.keras.Model):
    #
    # CONSTANTS
    #
    DEFAULT_KEY='steps'
    DEFAULTS=load.config(cfig='steps',key_path=DEFAULT_KEY)
    GAP='gap'
    SEGMENT='segment'



    #
    # STATIC
    #
    @staticmethod
    def from_config(
            cfig='steps',
            key_path=DEFAULT_KEY,
            cfig_dir=load.TFBOX,
            is_file_path=False,
            **kwargs):
        config=load.config(
            cfig=cfig,
            key_path=key_path,
            cfig_dir=cfig_dir,
            is_file_path=is_file_path,
            **kwargs)
        print('STEPS:')
        pprint(config)
        return Steps(**config)



    #
    # PUBLIC
    #
    def __init__(self,
            filters_list=DEFAULTS['filters_list'],
            strides_list=DEFAULTS.get('strides_list',1),
            kernel_size_list=DEFAULTS.get('kernel_size_list',3),
            depth_list=DEFAULTS.get('depth_list',1),
            dilation_rate_list=DEFAULTS.get('dilation_rate_list',1),
            squeeze_excitation_list=DEFAULTS.get('squeeze_excitation_list',False),
            residual_list=DEFAULTS.get('residual_list',False),
            nb_classes=DEFAULTS.get('nb_classes',None),
            classifier_type=DEFAULTS.get('classifier_type',SEGMENT),
            classifier_act=DEFAULTS.get('classifier_act',True),
            classifier_act_config=DEFAULTS.get('classifier_act_config',{}),
            classifier_kernel_size_list=DEFAULTS.get('classifier_kernel_size_list'),
            classifier_filters_list=DEFAULTS.get('classifier_filters_list'),
            name=DEFAULTS.get('name',None),
            named_layers=DEFAULTS.get('named_layers',True),
            **step_kwargs):
        super(Steps, self).__init__()
        self.model_name=name
        self.named_layers=named_layers
        self.nb_steps=len(filters_list)
        self.steps=self._steps(
            filters_list,
            self._as_list(strides_list),
            self._as_list(kernel_size_list),
            self._as_list(depth_list),
            self._as_list(dilation_rate_list),
            self._as_list(squeeze_excitation_list),
            self._as_list(residual_list),
            step_kwargs)
        if classifier_type==Steps.GAP:
            # todo: global-avg-pooling+dense+classifier
            pass
        elif classifier_type==Steps.SEGMENT:
            self.classifier=blocks.SegmentClassifier(
                nb_classes=nb_classes,
                filters_list=classifier_filters_list,
                kernel_size_list=classifier_kernel_size_list,
                output_act=classifier_act,
                output_act_config=classifier_act_config,
                name=self._layer_name('classifier'),
                named_layers=self.named_layers)
        else:
            self.classifier=False


    def __call__(self,x,training=False,**kwargs):
        for l in self.steps:
            x=l(x)
        if self.classifier:
            return self.classifier(x)
        else:
            return x


    #
    # INTERNAL
    #
    def _layer_name(self,group=None,index=None):
        return blocks.layer_name(
            *[self.model_name,group],
            index=index,
            named=self.named_layers)


    def _steps(self,
            filters_list,
            strides_list,
            kernel_size_list,
            depth_list,
            dilation_rate_list,
            squeeze_excitation_list,
            residual_list,
            step_config):
        _layers=[]
        for i,(f,s,k,depth,d,se,res) in enumerate(zip(
                filters_list,
                strides_list,
                kernel_size_list,
                depth_list,
                dilation_rate_list,
                squeeze_excitation_list,
                residual_list)):
            blk=blocks.Stack(
                filters=f,
                output_stride=s,
                kernel_size=k,
                depth=depth,
                dilation_rate=d,
                squeeze_excitation=se,
                residual=res,
                name=self._layer_name(group='step',index=i),
                named_layers=self.named_layers,
                **step_config)
            _layers.append(blk)
        return _layers


    def _as_list(self,value):
        if isinstance(value,int):
            value=[value]*self.nb_steps
        return value








from pprint import pprint
from . import load
from . import base
from . import blocks
#
# CONSTANTS
#
DEFAULT_BTYPE='stack'
INT_BTYPE='conv'
STACK='stack'


#
# Encoder: a flexible generic encoder
# 
class Encoder(base.Model):
    #
    # CONSTANTS
    #
    NAME='Encoder'


    def __init__(self,
            nb_classes=None,
            model_config=NAME,
            key_path='xception',
            is_file_path=False,
            cfig_dir=load.TFBOX,
            classifier_config={
                'classifier_type': base.Model.GLOBAL_POOLING,
            },
            return_empty_skips=False,
            name=NAME,
            named_layers=True,
            noisy=True):
        super(Encoder, self).__init__(
            name=name,
            named_layers=named_layers,
            noisy=noisy,
            nb_classes=nb_classes,
            classifier_config=classifier_config)
        if isinstance(model_config,str):
            model_config=load.config(
                    cfig=model_config,
                    key_path=key_path,
                    is_file_path=is_file_path,
                    cfig_dir=cfig_dir,
                    noisy=noisy )
        blocks_config=model_config['blocks_config']
        self.stacked_blocks=[self._stacked_blocks(c) for c in blocks_config]
        self.return_empty_skips=return_empty_skips


    def __call__(self,inputs,training=False):
        x=inputs
        skips=[]
        for sb in self.stacked_blocks:
            for block in sb:
                xin=x
                x=block(x)
                skips=self._update_skips(block,skips,x)
        if self.classifier:
            return self.classifier(x)
        elif skips or self.return_empty_skips:
            return x, skips
        else:
            return x


    #
    # INTERNAL
    #
    def _stacked_blocks(self,config):
        if isinstance(config,int):
            btype=INT_BTYPE
        elif isinstance(config,str):
            btype=config
            config=None
        else:
            nb_keys=len(config)
            if nb_keys==1:
                btype, config=list(iter(config.items()))[0]
            else:
                btype=DEFAULT_BTYPE
        return self._get_blocks(config,btype)


    def _get_blocks(self,config,btype):
        block_stack=[]
        if isinstance(config,dict):
            layers=config.pop('layers',None)
            nb_repeats=config.pop('nb_repeats',1)
        else:
            layers=False
            nb_repeats=False
        if layers or nb_repeats:
            if not layers:
                layers=[{}]
            for i in range(nb_repeats):
                for j,c in enumerate(layers):
                    cfg=config.copy()
                    cfg.update(self._as_config_dict(c))
                    block_stack.append(
                        self._get_block(cfg,btype,index=f'{i}-{j}'))
        else:
            block_stack.append(self._get_block(config,btype))
        return block_stack


    def _get_block(self,config,btype,index=None):
        config=self._as_config_dict(config)
        name=config.get('name')
        if index and name:
            config['name']=f'{name}.{index}'
        return blocks.get(btype)(**config)


    def _update_skips(self,block,skips,x,force_update=False):
        try:
            if force_update or (block.is_skip): 
                skips.append(x)
        except:
            pass
        return skips


    def _as_config_dict(self,config):
        if not config:
            config={}
        elif isinstance(config,int):
            config={'filters': config}
        return config












import os
import re
from pprint import pprint
import tfbox.utils.helpers as h
_nn_dir=os.path.dirname(os.path.realpath(__file__))
#
# CONSTANTS
#
TFBOX='tfbox'
CONFIGS_DIR=f'{_nn_dir}/configs'
YAML_RGX='.(yaml|yml)$'
JSON_RGX='.json$'





#
# I/0
#
def config(
        cfig=None,
        key_path=None,
        cfig_dir=TFBOX,
        is_file_path=False,
        config_string=None,
        noisy=False,
        **kwargs):
    """ load model configs
    Args:
        - cfig<str|dict|None>:
            * <str>: 
                - name (w/o '.yaml' ext) of config file
                - file_path (see is_file_path below)
            * <dict>: model config as dict
            * None: requires config_string
        - key_path<str>: 
            dot-path for key in cfig-file containing the config
        - cfig_dir<str>:
            - tfbox,None,True: use tfbox config file (tfbox.nn.configs)
            - False: current working directory
            - Otherwise: directory path
        - is_file_path<bool>:
            - if true: cfig arg is the path to the config file
        - config_string<str|None>:
            - if config_string derive above args using `parse_config_string`
        - kwargs: key-value args to update the loaded config
    """
    if config_string:
        cfig, key_path, is_file_path=parse_config_string(
            config_string,
            cfig=cfig)
    if isinstance(cfig,str):
        title=cfig
        cfig=h.snake(cfig)
        if not is_file_path:
            if cfig_dir in [TFBOX,None,True]:
                cfig_dir=CONFIGS_DIR
            if cfig_dir:
                cfig=f'{cfig_dir}/{cfig}'
        cfig=h.read_yaml(f'{cfig}.yaml')
    else:
        title='config'
    if key_path:
        key_path=h.snake(key_path)
        if isinstance(key_path,str):
            key_path=key_path.split('.')
        for k in key_path: cfig=cfig[k]
    if kwargs:
        if isinstance(cfig,dict):
            cfig.update(kwargs)
        else:
            raise ValueError('kwargs only allowed for key-value configs.')
    if noisy:
        pprint(cfig)
    return cfig



def parse_config_string(config_string,cfig=None):
    key_path=is_file_path=None
    parts=config_string.split(':')
    if cfig:
        parts=[cfig]+parts
    nb_parts=len(parts)
    cfig=parts[0]
    if nb_parts>1:
        key_path=parts[1]
        if nb_parts>2:
            is_file_path=str(parts[2]).lower()=='true'
    return cfig, key_path, is_file_path




